import { parse, format } from 'date-fns';
import * as _ from 'lodash';
import * as i0 from '@angular/core';
import { Injectable, NgModule } from '@angular/core';
import * as i1 from '@angular/common/http';
import { HttpErrorResponse, HttpHeaders, HttpClientModule } from '@angular/common/http';
import { map, catchError, mapTo } from 'rxjs/operators';
import { of, throwError } from 'rxjs';
import * as qs from 'qs';

const DEFAULT_OPTIONS = {
    nullValue: false,
    hasMany: false
};
class JsonModelConverter {
    constructor(model, options = {}) {
        this.modelType = model; // <ModelType<T>>model
        this.options = Object.assign(Object.assign({}, DEFAULT_OPTIONS), options);
    }
    mask(value) {
        if (!value && !this.options.nullValue) {
            if (this.options.hasMany) {
                return [];
            }
            return new this.modelType();
        }
        let result = null;
        if (this.options.hasMany) {
            if (!Array.isArray(value)) {
                throw new Error(`ERROR: JsonModelConverter: Expected array but got ${typeof value}.`);
            }
            result = [];
            for (const item of value) {
                if (item === null) {
                    continue;
                }
                let temp;
                if (typeof item === 'object') {
                    temp = new this.modelType();
                    temp.fill(item);
                }
                else {
                    temp = item;
                }
                result.push(temp);
            }
        }
        else {
            if (!(value instanceof this.modelType)) {
                result = new this.modelType();
                result.fill(value);
            }
            else {
                result = value;
            }
        }
        return result;
    }
    unmask(value) {
        if (!value) {
            return value;
        }
        let result = null;
        if (Array.isArray(value)) {
            result = [];
            for (const item of value) {
                if (!item) {
                    continue;
                }
                if (item instanceof JsonApiNestedModel) {
                    item.nestedDataSerialization = true;
                    result.push(item.serialize());
                    item.nestedDataSerialization = false;
                }
                else {
                    result.push(item);
                }
            }
        }
        else {
            if (value instanceof JsonApiNestedModel) {
                value.nestedDataSerialization = true;
                result = value.serialize();
                value.nestedDataSerialization = false;
            }
            else {
                result = value;
            }
        }
        return result;
    }
}

function HasMany(config = {}) {
    return function (target, propertyName) {
        const annotations = Reflect.getMetadata('HasMany', target) || [];
        annotations.push({
            propertyName,
            relationship: config.key || propertyName
        });
        Reflect.defineMetadata('HasMany', annotations, target);
    };
}

function BelongsTo(config = {}) {
    return function (target, propertyName) {
        const annotations = Reflect.getMetadata('BelongsTo', target) || [];
        annotations.push({
            propertyName,
            relationship: config.key || propertyName
        });
        Reflect.defineMetadata('BelongsTo', annotations, target);
    };
}

// tslint:disable-next-line:variable-name
const AttributeMetadata = Symbol('AttributeMetadata');

class DateConverter {
    mask(value) {
        if (value === undefined || value === null) {
            return undefined;
        }
        return parse(value);
    }
    unmask(value) {
        if (value === undefined || value === null) {
            return value;
        }
        return format(value, 'YYYY-MM-DDTHH:mm:ssZ');
    }
}

function Attribute(options = {}) {
    return function (target, propertyName) {
        const converter = function (dataType, value, forSerialisation = false) {
            let attrConverter;
            if (options.converter) {
                attrConverter = options.converter;
            }
            else if (dataType === Date) {
                attrConverter = new DateConverter();
            }
            else {
                const datatype = new dataType();
                if (datatype.mask && datatype.unmask) {
                    attrConverter = datatype;
                }
            }
            if (attrConverter) {
                if (!forSerialisation) {
                    return attrConverter.mask(value);
                }
                return attrConverter.unmask(value);
            }
            return value;
        };
        const saveAnnotations = function () {
            const metadata = Reflect.getMetadata('Attribute', target) || {};
            metadata[propertyName] = {
                marked: true
            };
            Reflect.defineMetadata('Attribute', metadata, target);
            const mappingMetadata = Reflect.getMetadata('AttributeMapping', target) || {};
            const serializedPropertyName = options.serializedName !== undefined ? options.serializedName : propertyName;
            mappingMetadata[serializedPropertyName] = propertyName;
            Reflect.defineMetadata('AttributeMapping', mappingMetadata, target);
        };
        const setMetadata = function (instance, oldValue, newValue) {
            const targetType = Reflect.getMetadata('design:type', target, propertyName);
            if (!instance[AttributeMetadata]) {
                instance[AttributeMetadata] = {};
            }
            instance[AttributeMetadata][propertyName] = {
                newValue,
                oldValue,
                nested: false,
                serializedName: options.serializedName,
                hasDirtyAttributes: !_.isEqual(oldValue, newValue),
                serialisationValue: converter(targetType, newValue, true)
            };
        };
        const getter = function () {
            return this[`_${propertyName}`];
        };
        const setter = function (newVal) {
            const targetType = Reflect.getMetadata('design:type', target, propertyName);
            const convertedValue = converter(targetType, newVal);
            let oldValue = null;
            if (this.isModelInitialization() && this.id) {
                oldValue = converter(targetType, newVal);
            }
            else {
                if (this[AttributeMetadata] && this[AttributeMetadata][propertyName]) {
                    oldValue = this[AttributeMetadata][propertyName]['oldValue'];
                }
            }
            this[`_${propertyName}`] = convertedValue;
            setMetadata(this, oldValue, convertedValue);
        };
        if (delete target[propertyName]) {
            saveAnnotations();
            Object.defineProperty(target, propertyName, {
                get: getter,
                set: setter,
                enumerable: true,
                configurable: true
            });
        }
    };
}

function NestedAttribute(options = {}) {
    return function (target, propertyName) {
        const converter = function (dataType, value, forSerialisation = false) {
            let attrConverter;
            if (options.converter) {
                attrConverter = options.converter;
            }
            else {
                const datatype = new dataType();
                if (datatype.mask && datatype.unmask) {
                    attrConverter = datatype;
                }
            }
            if (attrConverter) {
                if (!forSerialisation) {
                    return attrConverter.mask(value);
                }
                return attrConverter.unmask(value);
            }
            return value;
        };
        const saveAnnotations = function () {
            const metadata = Reflect.getMetadata('NestedAttribute', target) || {};
            metadata[propertyName] = {
                marked: true
            };
            Reflect.defineMetadata('NestedAttribute', metadata, target);
            const mappingMetadata = Reflect.getMetadata('AttributeMapping', target) || {};
            const serializedPropertyName = options.serializedName !== undefined ? options.serializedName : propertyName;
            mappingMetadata[serializedPropertyName] = propertyName;
            Reflect.defineMetadata('AttributeMapping', mappingMetadata, target);
        };
        const updateMetadata = (instance) => {
            const newValue = instance[`_${propertyName}`];
            if (!instance[AttributeMetadata]) {
                instance[AttributeMetadata] = {};
            }
            if (instance[AttributeMetadata][propertyName] && !instance.isModelInitialization()) {
                instance[AttributeMetadata][propertyName].newValue = newValue;
                instance[AttributeMetadata][propertyName].hasDirtyAttributes = !_.isEqual(instance[AttributeMetadata][propertyName].oldValue, newValue);
                instance[AttributeMetadata][propertyName].serialisationValue = newValue;
            }
            else {
                const oldValue = _.cloneDeep(newValue);
                instance[AttributeMetadata][propertyName] = {
                    newValue,
                    oldValue,
                    converter,
                    nested: true,
                    hasDirtyAttributes: !_.isEqual(newValue, oldValue)
                };
            }
        };
        const getter = function () {
            return this[`_${propertyName}`];
        };
        const setter = function (newVal) {
            const targetType = Reflect.getMetadata('design:type', target, propertyName);
            this[`_${propertyName}`] = converter(targetType, newVal);
            updateMetadata(this);
        };
        if (delete target[propertyName]) {
            saveAnnotations();
            Object.defineProperty(target, propertyName, {
                get: getter,
                set: setter,
                enumerable: true,
                configurable: true
            });
        }
    };
}

function JsonAttribute(options = {}) {
    return function (target, propertyName) {
        const converter = function (dataType, value, forSerialisation = false) {
            let attrConverter;
            if (options.converter) {
                attrConverter = options.converter;
            }
            else if (dataType === Date) {
                attrConverter = new DateConverter();
            }
            else {
                const datatype = new dataType();
                if (datatype.mask && datatype.unmask) {
                    attrConverter = datatype;
                }
            }
            if (attrConverter) {
                if (!forSerialisation) {
                    return attrConverter.mask(value);
                }
                return attrConverter.unmask(value);
            }
            return value;
        };
        const saveAnnotations = function () {
            const metadata = Reflect.getMetadata('JsonAttribute', target) || {};
            metadata[propertyName] = {
                marked: true
            };
            Reflect.defineMetadata('JsonAttribute', metadata, target);
            const mappingMetadata = Reflect.getMetadata('AttributeMapping', target) || {};
            const serializedPropertyName = options.serializedName !== undefined ? options.serializedName : propertyName;
            mappingMetadata[serializedPropertyName] = propertyName;
            Reflect.defineMetadata('AttributeMapping', mappingMetadata, target);
        };
        const getter = function () {
            if (this.nestedDataSerialization) {
                return converter(Reflect.getMetadata('design:type', target, propertyName), this[`_${propertyName}`], true);
            }
            return this[`_${propertyName}`];
        };
        const setter = function (newVal) {
            const targetType = Reflect.getMetadata('design:type', target, propertyName);
            this[`_${propertyName}`] = converter(targetType, newVal);
        };
        if (delete target[propertyName]) {
            saveAnnotations();
            Object.defineProperty(target, propertyName, {
                get: getter,
                set: setter,
                enumerable: true,
                configurable: true
            });
        }
    };
}

class JsonApiMetaModel {
    constructor(response) {
        this.links = response.links || [];
        this.meta = response.meta;
    }
}

function JsonApiModelConfig(config) {
    return function (target) {
        if (typeof config.meta === 'undefined' || config.meta == null) {
            config.meta = JsonApiMetaModel;
        }
        Reflect.defineMetadata('JsonApiModelConfig', config, target);
    };
}

function JsonApiDatastoreConfig(config = {}) {
    return function (target) {
        Reflect.defineMetadata('JsonApiDatastoreConfig', config, target);
    };
}

/**
 * HACK/FIXME:
 * Type 'symbol' cannot be used as an index type.
 * TypeScript 2.9.x
 * See https://github.com/Microsoft/TypeScript/issues/24587.
 */
// tslint:disable-next-line:variable-name
const AttributeMetadataIndex$1 = AttributeMetadata;
class JsonApiModel {
    constructor(internalDatastore, data) {
        this.internalDatastore = internalDatastore;
        this.modelInitialization = false;
        if (data) {
            this.modelInitialization = true;
            this.id = data.id;
            Object.assign(this, data.attributes);
            this.modelInitialization = false;
        }
    }
    isModelInitialization() {
        return this.modelInitialization;
    }
    syncRelationships(modelData, included, remainingModels) {
        if (this.lastSyncModels === included) {
            return;
        }
        if (modelData) {
            let modelsForProcessing = remainingModels;
            if (!modelsForProcessing) {
                modelsForProcessing = [].concat(included);
            }
            this.parseHasMany(modelData, included, modelsForProcessing);
            this.parseBelongsTo(modelData, included, modelsForProcessing);
        }
        this.lastSyncModels = included;
    }
    save(params, headers, customUrl) {
        this.checkChanges();
        const attributesMetadata = this[AttributeMetadataIndex$1];
        return this.internalDatastore.saveRecord(attributesMetadata, this, params, headers, customUrl);
    }
    get hasDirtyAttributes() {
        this.checkChanges();
        const attributesMetadata = this[AttributeMetadataIndex$1];
        let hasDirtyAttributes = false;
        for (const propertyName in attributesMetadata) {
            if (attributesMetadata.hasOwnProperty(propertyName)) {
                const metadata = attributesMetadata[propertyName];
                if (metadata.hasDirtyAttributes) {
                    hasDirtyAttributes = true;
                    break;
                }
            }
        }
        return hasDirtyAttributes;
    }
    checkChanges() {
        const attributesMetadata = this[AttributeMetadata];
        for (const propertyName in attributesMetadata) {
            if (attributesMetadata.hasOwnProperty(propertyName)) {
                const metadata = attributesMetadata[propertyName];
                if (metadata.nested) {
                    this[AttributeMetadata][propertyName].hasDirtyAttributes = !_.isEqual(attributesMetadata[propertyName].oldValue, attributesMetadata[propertyName].newValue);
                    this[AttributeMetadata][propertyName].serialisationValue = attributesMetadata[propertyName].converter(Reflect.getMetadata('design:type', this, propertyName), _.cloneDeep(attributesMetadata[propertyName].newValue), true);
                }
            }
        }
    }
    rollbackAttributes() {
        const attributesMetadata = this[AttributeMetadataIndex$1];
        for (const propertyName in attributesMetadata) {
            if (attributesMetadata.hasOwnProperty(propertyName)) {
                if (attributesMetadata[propertyName].hasDirtyAttributes) {
                    this[propertyName] = _.cloneDeep(attributesMetadata[propertyName].oldValue);
                }
            }
        }
    }
    get modelConfig() {
        return Reflect.getMetadata('JsonApiModelConfig', this.constructor);
    }
    parseHasMany(modelData, included, remainingModels) {
        const hasMany = Reflect.getMetadata('HasMany', this);
        if (!hasMany) {
            return;
        }
        for (const metadata of hasMany) {
            const relationship = modelData.relationships ? modelData.relationships[metadata.relationship] : null;
            if (!(relationship && relationship.data && relationship.data.length > 0)) {
                continue;
            }
            let allModels = [];
            const modelTypesFetched = [];
            for (const typeIndex of Object.keys(relationship.data)) {
                const typeName = relationship.data[typeIndex].type;
                if (!_.includes(modelTypesFetched, typeName)) {
                    modelTypesFetched.push(typeName);
                    // tslint:disable-next-line:max-line-length
                    const modelType = Reflect.getMetadata('JsonApiDatastoreConfig', this.internalDatastore.constructor).models[typeName];
                    if (!modelType) {
                        throw { message: `parseHasMany - Model type for relationship ${typeName} not found.` };
                    }
                    const relationshipModels = this.getHasManyRelationship(modelType, relationship.data, included, typeName, remainingModels);
                    if (relationshipModels.length > 0) {
                        allModels = allModels.concat(relationshipModels);
                    }
                }
                if (allModels.length > 0) {
                    this[metadata.propertyName] = allModels;
                }
            }
        }
    }
    parseBelongsTo(data, included, remainingModels) {
        const belongsTo = Reflect.getMetadata('BelongsTo', this);
        if (!belongsTo) {
            return;
        }
        for (const metadata of belongsTo) {
            const relationship = data.relationships ? data.relationships[metadata.relationship] : null;
            if (relationship && relationship.data) {
                const dataRelationship = (relationship.data instanceof Array) ? relationship.data[0] : relationship.data;
                if (dataRelationship) {
                    const typeName = dataRelationship.type;
                    // tslint:disable-next-line:max-line-length
                    const modelType = Reflect.getMetadata('JsonApiDatastoreConfig', this.internalDatastore.constructor).models[typeName];
                    if (modelType) {
                        const relationshipModel = this.getBelongsToRelationship(modelType, dataRelationship, included, typeName, remainingModels);
                        if (relationshipModel) {
                            this[metadata.propertyName] = relationshipModel;
                        }
                    }
                    else {
                        throw { message: `parseBelongsTo - Model type for relationship ${typeName} not found.` };
                    }
                }
            }
        }
    }
    getHasManyRelationship(modelType, data, included, typeName, remainingModels) {
        const relationshipList = [];
        data.forEach((item) => {
            const relationshipData = _.chain(remainingModels)
                .filter(['id', item.id])
                .filter(['type', typeName])
                .first()
                .value();
            if (relationshipData) {
                const newObject = this.createOrPeek(modelType, relationshipData);
                const indexOfNewlyFoundModel = remainingModels.indexOf(relationshipData);
                const modelsForProcessing = remainingModels.concat([]);
                modelsForProcessing.splice(indexOfNewlyFoundModel, 1);
                newObject.syncRelationships(relationshipData, included, modelsForProcessing);
                relationshipList.push(newObject);
            }
        });
        return relationshipList;
    }
    getBelongsToRelationship(modelType, data, included, typeName, remainingModels) {
        const id = data.id;
        const relationshipData = _.chain(remainingModels)
            .filter(['id', id])
            .filter(['type', typeName])
            .first()
            .value();
        if (relationshipData) {
            const newObject = this.createOrPeek(modelType, relationshipData);
            const indexOfNewlyFoundModel = remainingModels.indexOf(relationshipData);
            const modelsForProcessing = remainingModels.concat([]);
            modelsForProcessing.splice(indexOfNewlyFoundModel, 1);
            newObject.syncRelationships(relationshipData, included, modelsForProcessing);
            return newObject;
        }
        return this.internalDatastore.peekRecord(modelType, id);
    }
    createOrPeek(modelType, data) {
        const peek = this.internalDatastore.peekRecord(modelType, data.id);
        if (peek) {
            _.extend(peek, this.internalDatastore.transformSerializedNamesToPropertyNames(modelType, data.attributes));
            return peek;
        }
        const newObject = this.internalDatastore.deserializeModel(modelType, data);
        this.internalDatastore.addToStore(newObject);
        return newObject;
    }
}

class JsonApiNestedModel {
    constructor(data) {
        this.nestedDataSerialization = false;
        if (data) {
            Object.assign(this, data);
        }
    }
    get modelConfig() {
        return Reflect.getMetadata('JsonApiModelConfig', this.constructor);
    }
    fill(data) {
        Object.assign(this, data);
    }
    serialize() {
        return this.transformSerializedNamesToPropertyNames();
    }
    transformSerializedNamesToPropertyNames() {
        const serializedNameToPropertyName = this.getModelPropertyNames();
        const properties = {};
        Object.keys(serializedNameToPropertyName).forEach((serializedName) => {
            if (this && this[serializedName] !== null &&
                this[serializedName] !== undefined && serializedName !== 'nestedDataSerialization') {
                properties[serializedNameToPropertyName[serializedName]] = this[serializedName];
            }
        });
        return properties;
    }
    getModelPropertyNames() {
        return Reflect.getMetadata('AttributeMapping', this) || [];
    }
}

class ErrorResponse {
    constructor(errors) {
        this.errors = [];
        if (errors) {
            this.errors = errors;
        }
    }
}

class JsonApiQueryData {
    constructor(jsonApiModels, metaData) {
        this.jsonApiModels = jsonApiModels;
        this.metaData = metaData;
    }
    getModels() {
        return this.jsonApiModels;
    }
    getMeta() {
        return this.metaData;
    }
}

/**
 * HACK/FIXME:
 * Type 'symbol' cannot be used as an index type.
 * TypeScript 2.9.x
 * See https://github.com/Microsoft/TypeScript/issues/24587.
 */
// tslint:disable-next-line:variable-name
const AttributeMetadataIndex = AttributeMetadata;
class JsonApiDatastore {
    get getDirtyAttributes() {
        if (this.datastoreConfig.overrides
            && this.datastoreConfig.overrides.getDirtyAttributes) {
            return this.datastoreConfig.overrides.getDirtyAttributes;
        }
        return JsonApiDatastore.getDirtyAttributes;
    }
    constructor(http) {
        this.http = http;
        this.globalRequestOptions = {};
        this.internalStore = {};
        this.toQueryString = this.datastoreConfig.overrides
            && this.datastoreConfig.overrides.toQueryString ?
            this.datastoreConfig.overrides.toQueryString : this._toQueryString;
    }
    /** @deprecated - use findAll method to take all models **/
    query(modelType, params, headers, customUrl) {
        const requestHeaders = this.buildHttpHeaders(headers);
        const url = this.buildUrl(modelType, params, undefined, customUrl);
        return this.http.get(url, { headers: requestHeaders })
            .pipe(map((res) => this.extractQueryData(res, modelType)), catchError((res) => this.handleError(res)));
    }
    findAll(modelType, params, headers, customUrl) {
        const url = this.buildUrl(modelType, params, undefined, customUrl);
        const requestOptions = this.buildRequestOptions({ headers });
        return this.http.get(url, requestOptions)
            .pipe(map((res) => this.extractQueryData(res, modelType, true)), catchError((res) => this.handleError(res)));
    }
    findRecord(modelType, id, params, headers, customUrl) {
        const requestOptions = this.buildRequestOptions({ headers, observe: 'response' });
        const url = this.buildUrl(modelType, params, id, customUrl);
        return this.http.get(url, requestOptions)
            .pipe(map((res) => this.extractRecordData(res, modelType)), catchError((res) => this.handleError(res)));
    }
    createRecord(modelType, data) {
        return new modelType(this, { attributes: data });
    }
    static getDirtyAttributes(attributesMetadata) {
        const dirtyData = {};
        for (const propertyName in attributesMetadata) {
            if (attributesMetadata.hasOwnProperty(propertyName)) {
                const metadata = attributesMetadata[propertyName];
                if (metadata.hasDirtyAttributes) {
                    const attributeName = metadata.serializedName != null ? metadata.serializedName : propertyName;
                    dirtyData[attributeName] = metadata.serialisationValue ? metadata.serialisationValue : metadata.newValue;
                }
            }
        }
        return dirtyData;
    }
    saveRecord(attributesMetadata, model, params, headers, customUrl) {
        const modelType = model.constructor;
        const modelConfig = model.modelConfig;
        const typeName = modelConfig.type;
        const requestHeaders = this.buildHttpHeaders(headers);
        const relationships = this.getRelationships(model);
        const url = this.buildUrl(modelType, params, model.id, customUrl);
        let httpCall;
        const body = {
            data: {
                relationships,
                type: typeName,
                id: model.id,
                attributes: this.getDirtyAttributes(attributesMetadata, model)
            }
        };
        const requestOptions = this.buildRequestOptions({ headers, observe: 'response' });
        if (model.id) {
            httpCall = this.http.patch(url, body, requestOptions);
        }
        else {
            httpCall = this.http.post(url, body, requestOptions);
        }
        return httpCall
            .pipe(map((res) => [200, 201].indexOf(res.status) !== -1 ? this.extractRecordData(res, modelType, model) : model), catchError((res) => {
            if (res == null) {
                return of(model);
            }
            return this.handleError(res);
        }), map((res) => this.updateRelationships(res, relationships)));
    }
    deleteRecord(modelType, id, headers, customUrl) {
        const requestOptions = this.buildRequestOptions({ headers });
        const url = this.buildUrl(modelType, null, id, customUrl);
        return this.http.delete(url, requestOptions)
            .pipe(catchError((res) => this.handleError(res)), mapTo(undefined));
    }
    peekRecord(modelType, id) {
        const type = Reflect.getMetadata('JsonApiModelConfig', modelType).type;
        return this.internalStore[type] ? this.internalStore[type][id] : null;
    }
    peekAll(modelType) {
        const type = Reflect.getMetadata('JsonApiModelConfig', modelType).type;
        const typeStore = this.internalStore[type];
        return typeStore ? Object.keys(typeStore).map((key) => typeStore[key]) : [];
    }
    set headers(headers) {
        this.globalHeaders = headers;
    }
    set requestOptions(requestOptions) {
        this.globalRequestOptions = requestOptions;
    }
    buildUrl(modelType, params, id, customUrl) {
        // TODO: use HttpParams instead of appending a string to the url
        const queryParams = this.toQueryString(params);
        if (customUrl) {
            return queryParams ? `${customUrl}?${queryParams}` : customUrl;
        }
        const modelConfig = Reflect.getMetadata('JsonApiModelConfig', modelType);
        const baseUrl = modelConfig.baseUrl || this.datastoreConfig.baseUrl;
        const apiVersion = modelConfig.apiVersion || this.datastoreConfig.apiVersion;
        const modelEndpointUrl = modelConfig.modelEndpointUrl || modelConfig.type;
        const url = [baseUrl, apiVersion, modelEndpointUrl, id].filter((x) => x).join('/');
        return queryParams ? `${url}?${queryParams}` : url;
    }
    getRelationships(model) {
        let relationships;
        const data = model;
        _.forOwn(data, (propertyValue, propertyName) => {
            if (!(propertyValue instanceof JsonApiModel)) {
                return;
            }
            relationships = relationships || {};
            if (!propertyValue.id) {
                return;
            }
            relationships[propertyName] = {
                data: this.buildSingleRelationshipData(propertyValue)
            };
        });
        _.forOwn(data, (propertyValue, propertyName) => {
            const hasMany = propertyValue instanceof Array &&
                propertyValue.length > 0 &&
                this.isValidToManyRelation(propertyValue);
            if (!hasMany) {
                return;
            }
            relationships = relationships || {};
            const relationshipData = data[propertyName]
                .filter((model) => model.id)
                .map((model) => this.buildSingleRelationshipData(model));
            relationships[propertyName] = {
                data: relationshipData
            };
        });
        this.belongsToRelationsToBeDeleted(model).forEach((toDelete) => {
            relationships = relationships || {};
            relationships[toDelete] = { data: null };
        });
        this.hasManyRelationsToBeDeleted(model).forEach((toDelete) => {
            relationships = relationships || {};
            relationships[toDelete] = { data: [] };
        });
        return relationships;
    }
    belongsToRelationsToBeDeleted(model) {
        const belongsToMetadata = Reflect.getMetadata('BelongsTo', model) || [];
        return belongsToMetadata.filter((entity) => model.hasOwnProperty(entity.propertyName))
            .filter((entity) => model[entity.propertyName] === null)
            .map((entity) => entity.relationship);
    }
    hasManyRelationsToBeDeleted(model) {
        const hasManyMetadata = Reflect.getMetadata('HasMany', model) || [];
        return hasManyMetadata.filter((entity) => model.hasOwnProperty(entity.propertyName))
            .filter((entity) => _.isEmpty(model[entity.propertyName]))
            .map((entity) => entity.relationship);
    }
    isValidToManyRelation(objects) {
        const isJsonApiModel = objects.every((item) => item instanceof JsonApiModel);
        const relationshipType = isJsonApiModel ? objects[0].modelConfig.type : '';
        return isJsonApiModel ? objects.every((item) => item.modelConfig.type === relationshipType) : false;
    }
    buildSingleRelationshipData(model) {
        const relationshipType = model.modelConfig.type;
        const relationshipData = { type: relationshipType };
        if (model.id) {
            relationshipData.id = model.id;
        }
        else {
            const attributesMetadata = Reflect.getMetadata('Attribute', model);
            relationshipData.attributes = this.getDirtyAttributes(attributesMetadata, model);
        }
        return relationshipData;
    }
    extractQueryData(response, modelType, withMeta = false) {
        const unwrapedResponse = response;
        const data = unwrapedResponse.data;
        const included = unwrapedResponse.included;
        const models = [];
        data.forEach((modelData) => {
            const model = this.deserializeModel(modelType, modelData);
            this.addToStore(model);
            model.syncRelationships(modelData, data);
            this.addToStore(model);
            if (included) {
                model.syncRelationships(modelData, included);
                this.addToStore(model);
            }
            models.push(model);
        });
        if (withMeta && withMeta === true) {
            return new JsonApiQueryData(models, this.parseMeta(response, modelType));
        }
        return models;
    }
    deserializeModel(modelType, data) {
        data.attributes = this.transformSerializedNamesToPropertyNames(modelType, data.attributes);
        return new modelType(this, data);
    }
    extractRecordData(res, modelType, model) {
        const body = res.body;
        // Error in Angular < 5.2.4 (see https://github.com/angular/angular/issues/20744)
        // null is converted to 'null', so this is temporary needed to make testcase possible
        // (and to avoid a decrease of the coverage)
        if (!body || body === 'null') {
            throw new Error('no body in response');
        }
        if (!body.data) {
            if (res.status === 201 || !model) {
                throw new Error('expected data in response');
            }
            return model;
        }
        if (model) {
            model.modelInitialization = true;
            model.id = body.data.id;
            Object.assign(model, body.data.attributes);
            model.modelInitialization = false;
        }
        const deserializedModel = model || this.deserializeModel(modelType, body.data);
        this.addToStore(deserializedModel);
        if (body.included) {
            deserializedModel.syncRelationships(body.data, body.included);
            this.addToStore(deserializedModel);
        }
        return deserializedModel;
    }
    handleError(error) {
        if (error instanceof HttpErrorResponse &&
            error.error instanceof Object &&
            error.error.errors &&
            error.error.errors instanceof Array) {
            const errors = new ErrorResponse(error.error.errors);
            return throwError(errors);
        }
        return throwError(error);
    }
    parseMeta(body, modelType) {
        const metaModel = Reflect.getMetadata('JsonApiModelConfig', modelType).meta;
        return new metaModel(body);
    }
    /** @deprecated - use buildHttpHeaders method to build request headers **/
    getOptions(customHeaders) {
        return {
            headers: this.buildHttpHeaders(customHeaders),
        };
    }
    buildHttpHeaders(customHeaders) {
        let requestHeaders = new HttpHeaders({
            Accept: 'application/vnd.api+json',
            'Content-Type': 'application/vnd.api+json'
        });
        requestHeaders = this.merge(requestHeaders, this.globalHeaders);
        requestHeaders = this.merge(requestHeaders, customHeaders);
        return requestHeaders;
    }
    merge(a, b) {
        if (!b) {
            return a;
        }
        return _.chain(b.keys())
            .filter((key) => b.has(key))
            .reduce((a, key) => a.set(key, b.get(key)), a)
            .value();
    }
    buildRequestOptions(customOptions = {}) {
        const httpHeaders = this.buildHttpHeaders(customOptions.headers);
        const requestOptions = Object.assign(customOptions, {
            headers: httpHeaders
        });
        return Object.assign(this.globalRequestOptions, requestOptions);
    }
    _toQueryString(params) {
        return qs.stringify(params, { arrayFormat: 'brackets' });
    }
    addToStore(modelOrModels) {
        const models = Array.isArray(modelOrModels) ? modelOrModels : [modelOrModels];
        const type = models[0].modelConfig.type;
        let typeStore = this.internalStore[type];
        if (!typeStore) {
            typeStore = this.internalStore[type] = {};
        }
        for (const model of models) {
            typeStore[model.id] = model;
        }
    }
    resetMetadataAttributes(res, attributesMetadata, modelType) {
        for (const propertyName in attributesMetadata) {
            if (!attributesMetadata.hasOwnProperty(propertyName)) {
                continue;
            }
            const metadata = attributesMetadata[propertyName];
            if (metadata.hasDirtyAttributes) {
                metadata.hasDirtyAttributes = false;
            }
        }
        res[AttributeMetadataIndex] = attributesMetadata;
        return res;
    }
    updateRelationships(model, relationships) {
        const modelsTypes = Reflect.getMetadata('JsonApiDatastoreConfig', this.constructor).models;
        for (const relationship in relationships) {
            if (!relationships.hasOwnProperty(relationship) || !model.hasOwnProperty(relationship)) {
                continue;
            }
            const relationshipModel = model[relationship];
            const hasMany = Reflect.getMetadata('HasMany', relationshipModel);
            const propertyHasMany = _.find(hasMany, (property) => {
                return modelsTypes[property.relationship] === model.constructor;
            });
            if (!propertyHasMany) {
                continue;
            }
            relationshipModel[propertyHasMany.propertyName] = relationshipModel[propertyHasMany.propertyName] || [];
            const indexOfModel = relationshipModel[propertyHasMany.propertyName].indexOf(model);
            if (indexOfModel === -1) {
                relationshipModel[propertyHasMany.propertyName].push(model);
            }
            else {
                relationshipModel[propertyHasMany.propertyName][indexOfModel] = model;
            }
        }
        return model;
    }
    get datastoreConfig() {
        const configFromDecorator = Reflect.getMetadata('JsonApiDatastoreConfig', this.constructor);
        return Object.assign(configFromDecorator, this.config);
    }
    transformSerializedNamesToPropertyNames(modelType, attributes) {
        const serializedNameToPropertyName = this.getModelPropertyNames(modelType.prototype);
        const properties = {};
        Object.keys(serializedNameToPropertyName).forEach((serializedName) => {
            if (attributes && attributes[serializedName] !== null && attributes[serializedName] !== undefined) {
                properties[serializedNameToPropertyName[serializedName]] = attributes[serializedName];
            }
        });
        return properties;
    }
    getModelPropertyNames(model) {
        return Reflect.getMetadata('AttributeMapping', model) || [];
    }
}
/** @nocollapse */ JsonApiDatastore.ɵfac = function JsonApiDatastore_Factory(t) { return new (t || JsonApiDatastore)(i0.ɵɵinject(i1.HttpClient)); };
/** @nocollapse */ JsonApiDatastore.ɵprov = /** @pureOrBreakMyCode */ i0.ɵɵdefineInjectable({ token: JsonApiDatastore, factory: JsonApiDatastore.ɵfac });
(function () {
    (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(JsonApiDatastore, [{
            type: Injectable
        }], function () { return [{ type: i1.HttpClient }]; }, null);
})();

const PROVIDERS = [
    JsonApiDatastore
];

class JsonApiModule {
}
/** @nocollapse */ JsonApiModule.ɵfac = function JsonApiModule_Factory(t) { return new (t || JsonApiModule)(); };
/** @nocollapse */ JsonApiModule.ɵmod = /** @pureOrBreakMyCode */ i0.ɵɵdefineNgModule({ type: JsonApiModule });
/** @nocollapse */ JsonApiModule.ɵinj = /** @pureOrBreakMyCode */ i0.ɵɵdefineInjector({ providers: [PROVIDERS], imports: [HttpClientModule] });
(function () {
    (typeof ngDevMode === "undefined" || ngDevMode) && i0.ɵsetClassMetadata(JsonApiModule, [{
            type: NgModule,
            args: [{
                    providers: [PROVIDERS],
                    exports: [HttpClientModule]
                }]
        }], null, null);
})();
(function () { (typeof ngJitMode === "undefined" || ngJitMode) && i0.ɵɵsetNgModuleScope(JsonApiModule, { exports: [HttpClientModule] }); })();

/*
 * Public API Surface of angular2-jsonapi
 */

/**
 * Generated bundle index. Do not edit.
 */

export { Attribute, BelongsTo, DEFAULT_OPTIONS, ErrorResponse, HasMany, JsonApiDatastore, JsonApiDatastoreConfig, JsonApiMetaModel, JsonApiModel, JsonApiModelConfig, JsonApiModule, JsonApiNestedModel, JsonApiQueryData, JsonAttribute, JsonModelConverter, NestedAttribute, PROVIDERS };
//# sourceMappingURL=rvrvrv-angular2-jsonapi.mjs.map
